import { Keypair } from '@solana/web3.js';
import test from 'tape';
import {
  assertThrows,
  createWallet,
  killStuckProcess,
  metaplex,
} from '../../../helpers';
import { assertMintingWasSuccessful, createCandyMachine } from '../helpers';
import { isEqualToAmount, sol, toBigNumber } from '@/index';

killStuckProcess();

test('[candyMachineModule] addressGate guard: it allows minting from a specific address only', async (t) => {
  // Given a loaded Candy Machine with an addressGate guard.
  const mx = await metaplex();
  const allowedAddress = await createWallet(mx, 10);
  const { candyMachine, collection } = await createCandyMachine(mx, {
    itemsAvailable: toBigNumber(1),
    items: [{ name: 'Degen #1', uri: 'https://example.com/degen/1' }],
    guards: {
      addressGate: {
        address: allowedAddress.publicKey,
      },
    },
  });

  // When the allowed address mints from it.
  const { nft } = await mx.candyMachines().mint(
    {
      candyMachine,
      collectionUpdateAuthority: collection.updateAuthority.publicKey,
    },
    { payer: allowedAddress }
  );

  // Then minting was successful.
  await assertMintingWasSuccessful(t, mx, {
    candyMachine,
    collectionUpdateAuthority: collection.updateAuthority.publicKey,
    nft,
    owner: allowedAddress.publicKey,
  });
});

test('[candyMachineModule] addressGate guard: it forbids minting from anyone else', async (t) => {
  // Given a loaded Candy Machine with an addressGate guard.
  const mx = await metaplex();
  const allowedAddress = Keypair.generate();
  const { candyMachine, collection } = await createCandyMachine(mx, {
    itemsAvailable: toBigNumber(1),
    items: [{ name: 'Degen #1', uri: 'https://example.com/degen/1' }],
    guards: {
      addressGate: {
        address: allowedAddress.publicKey,
      },
    },
  });

  // When the another wallet tries to mint from it.
  const payer = await createWallet(mx, 10);
  const promise = mx.candyMachines().mint(
    {
      candyMachine,
      collectionUpdateAuthority: collection.updateAuthority.publicKey,
    },
    { payer }
  );

  // Then we expect an error.
  await assertThrows(t, promise, /Address not authorized/);
});

test('[candyMachineModule] addressGate guard with bot tax: it charges a bot tax when trying to mint using the wrong address', async (t) => {
  // Given a loaded Candy Machine with an addressGate guard and a bot tax guard.
  const mx = await metaplex();
  const allowedAddress = Keypair.generate();
  const { candyMachine, collection } = await createCandyMachine(mx, {
    itemsAvailable: toBigNumber(1),
    items: [{ name: 'Degen #1', uri: 'https://example.com/degen/1' }],
    guards: {
      botTax: {
        lamports: sol(0.1),
        lastInstruction: true,
      },
      addressGate: {
        address: allowedAddress.publicKey,
      },
    },
  });

  // When the another wallet tries to mint from it.
  const payer = await createWallet(mx, 10);
  const promise = mx.candyMachines().mint(
    {
      candyMachine,
      collectionUpdateAuthority: collection.updateAuthority.publicKey,
    },
    { payer }
  );

  // Then we expect a bot tax error.
  await assertThrows(t, promise, /CandyMachineBotTaxError/);

  // And the payer was charged a bot tax.
  const payerBalance = await mx.rpc().getBalance(payer.publicKey);
  t.true(
    isEqualToAmount(payerBalance, sol(9.9), sol(0.01)),
    'payer was charged a bot tax'
  );
});
